""" Exceptions for Whistle Client """

from typing import Any

class WhistleAuthError(Exception):
    """ Authentication Error """

    def __init__(self, *args: Any) -> None:
        """Initialize the exception."""
        Exception.__init__(self, *args)

class WhistleError(Exception):
    """Whistle error exception."""

    def __init__(self, *args: Any) -> None:
        """Initialize the exception."""
        Exception.__init__(self, *args)



""" Data classes for Whistle API """
from __future__ import annotations

from dataclasses import dataclass
from typing import Any



@dataclass
class WhistleData:
    """ Dataclass for Whistle data. """

    pets: dict[str, Pet]



@dataclass
class Pet:
    """ Dataclass for Whistle Pet. """

    id: str
    data: dict[str, Any]
    device: dict[str, Any]
    dailies: dict[str, Any]
    events: dict[str, Any] | None
    places: list[dict]
    stats: dict[str, Any]
    health: dict[str, Any]




""" Python client for Whistle API """
from __future__ import annotations

from typing import Any

import asyncio
from aiohttp import ClientResponse, ClientSession

from whistleaio.const import Endpoint, Header, TIMEOUT

from whistleaio.exceptions import WhistleAuthError, WhistleError

from whistleaio.model import WhistleData, Pet


class WhistleClient:
    """ Whistle Client. """

    def __init__(
            self, email: str, password: str,
            session: ClientSession | None = None,
            timeout: int = TIMEOUT
    ) -> None:
        """
        email: Registered whistle account email
        password: Registered whistle account password
        session: aiohttp.ClientSession or None to create a new session
        """

        self.email: str = email
        self.password: str = password
        self._session = session if session else ClientSession()
        self.token: str | None = None
        self.timeout: int = timeout

    async def get_token(self) -> None:
        """ Get auth token. No header needed. """

        data = {
            "email": self.email,
            "password": self.password
        }

        response = await self._post(endpoint=Endpoint.LOGIN, header={}, data=data)
        self.token = response['auth_token']

    async def get_whistle_data(self) -> WhistleData:
        """Fetch info for all pets and devices associated with
        Whistle account.
        """

        pets_data: dict[str, Pet] = {}
        response = await self.get_pets()

        if response['pets']:
            for pet in response['pets']:
                all_endpoints = await self.fetch_all_endpoints(pet)
                daily_item = await self.get_dailies_daily_items(pet['id'],
                                                                all_endpoints[1]['dailies'][00]['day_number'])

                pets_data[str(pet['id'])] = Pet(
                    id=str(pet['id']),
                    data=pet,
                    device=all_endpoints[0],
                    dailies=all_endpoints[1],
                    events=daily_item,
                    places=all_endpoints[2],
                    stats=all_endpoints[3],
                    health=all_endpoints[4]
                )
        return WhistleData(pets=pets_data)

    async def get_pets(self) -> dict[str, Any]:
        """ Get all pets. """

        if self.token is None:
            await self.get_token()
        header = await self.create_header()
        response = await self._get(endpoint=Endpoint.PETS, header=header)
        return response

    async def get_device_data(self, device_serial: str) -> dict[str, Any]:
        """ Get data for a single device. """

        if self.token is None:
            await self.get_token()
        header = await self.create_header()
        endpoint = f'{Endpoint.DEVICES}/{device_serial}'
        response = await self._get(endpoint=endpoint, header=header)
        return response

    async def get_dailies(self, pet_id: int) -> dict[str, Any]:
        """ Get dailies data for single pet. """

        if self.token is None:
            await self.get_token()
        header = await self.create_header()
        endpoint = f'{Endpoint.PETS}/{pet_id}{Endpoint.DAILIES}'
        response = await self._get(endpoint=endpoint, header=header)
        return response

    async def get_dailies_daily_items(self, pet_id: int, day_number: int) -> dict[str, Any]:
        """Get dailies daily items for single pet. The events reside
        at this endpoint.
        """

        if self.token is None:
            await self.get_token()
        header = await self.create_header()
        endpoint = f'{Endpoint.PETS}/{pet_id}{Endpoint.DAILIES}/{day_number}/daily_items'
        response = await self._get(endpoint=endpoint, header=header)
        return response

    async def get_stats(self, pet_id: int) -> dict[str, Any]:
        """ Get stats for single pet. """

        if self.token is None:
            await self.get_token()
        header = await self.create_header()
        endpoint = f'{Endpoint.PETS}/{pet_id}{Endpoint.STATS}'
        response = await self._get(endpoint=endpoint, header=header)
        return response

    async def get_places(self) -> dict[str, Any]:
        """ Get all places created within Whistle app. """

        if self.token is None:
            await self.get_token()
        header = await self.create_header()
        response = await self._get(endpoint=Endpoint.PLACES, header=header)
        return response

    async def get_health_trends(self, pet_id: int) -> dict[str, Any]:
        """Get all the health trends associated with a pet."""

        if self.token is None:
            await self.get_token()
        header = await self.create_header()
        endpoint = f'{Endpoint.PETS}/{pet_id}{Endpoint.HEALTH}'
        response = await self._get(endpoint=endpoint, header=header)
        return response

    async def fetch_all_endpoints(self, pet: dict[str, Any]) -> list:
        """Parallel request are made to all endpoints needed to
        get data for device, dailies, places, and stats of the Pet Object.
        """

        results = await asyncio.gather(*[
            self.get_device_data(pet['device']['serial_number']),
            self.get_dailies(pet['id']),
            self.get_places(),
            self.get_stats(pet['id']),
            self.get_health_trends(pet['id'])
            ],
        )
        return results

    async def create_header(self) -> dict[str, str]:
        """ Creates header that is used in all calls except token retrieval. """

        header = {
            "Accept": Header.ACCEPT,
            "Accept-Encoding": Header.ACCEPT_ENCODING,
            "Accept-Language": Header.LANGUAGE,
            "Accept-Unit-System": Header.UNIT,
            "Connection": 'keep-alive',
            "Content-Type": Header.CONTENT_TYPE,
            "User-Agent": Header.AGENT,
            "Authorization": f'Bearer {self.token}'
        }
        return header

    async def _post(self, endpoint: str, header: dict[str, Any], data: dict[str, Any]) -> dict[str, Any]:
        """ Make POST call to Whistle servers. """

        async with self._session.post(
            url=f'{Endpoint.BASE_URL}{endpoint}', headers=header,
                data=data, timeout=self.timeout) as resp:
            return await self._response(resp)

    async def _get(self, endpoint: str, header: dict[str, Any]) -> dict[str, Any]:
        """ Make GET call to Whistle servers. """

        async with self._session.get(
            url=f'{Endpoint.BASE_URL}{endpoint}', headers=header,
                timeout=self.timeout) as resp:
            return await self._response(resp)

    @staticmethod
    async def _response(resp: ClientResponse) -> dict[str, Any] | None:
        """ Check response for any errors & return original response if none """

        try:
            response: dict[str, Any] = await resp.json()
        except Exception as ex:
            raise WhistleError(ex)
        if resp.status == 422:
            if response['errors'][0]['message'] == 'Invalid email address or password':
                raise WhistleAuthError('Invalid email address or password')
        else:
            return response




from strenum import StrEnum

TIMEOUT = 5 * 60

""" Endpoint URLs """

class Endpoint(StrEnum):
    """ Endpoint type for Whistle API. """

    BASE_URL = 'https://app.whistle.com/api'
    DAILIES = '/dailies'
    DEVICES = '/devices'
    HEALTH = '/health/trends'
    LOGIN = '/login'
    PETS = '/pets'
    PLACES= '/places'
    STATS = '/stats'


""" Header """

class Header(StrEnum):
    """ Header type for Whistle API. """

    ACCEPT = 'application/vnd.whistle.com.v6+json'
    ACCEPT_ENCODING = 'br;q=1.0, gzip;q=0.9, deflate;q=0.8'
    AGENT = 'Winston/5.4.0 (iPhone; iOS 15.1; Build:3821; Scale/3.0'
    CONTENT_TYPE = 'application/json'
    LANGUAGE = 'en'
    UNIT = 'imperial'








import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="whistleaio",
    version="0.1.1",
    author="Robert Drinovac",
    author_email="unlisted@gmail.com",
    description="Asynchronous Python library for Whistle's API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/RobertD502/whistleaio',
    keywords='whistle, whistle fit, whistle api, whistle client, whistle gps, whistle pet',
    packages=setuptools.find_packages(),
    python_requires= ">=3.7",
    install_requires=[
        "aiohttp>=3.8.1",
        "strenum>=0.4.8",
    ],
    classifiers=(
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ),
    project_urls={  # Optional
    'Bug Reports': 'https://github.com/RobertD502/whistleaio/issues',
    'Source': 'https://github.com/RobertD502/whistleaio/',
    },
)



from .client import (WhistleClient,)
from .const import (Endpoint, Header, TIMEOUT,)
from .exceptions import (WhistleAuthError, WhistleError)
from .model import (Pet, WhistleData,)

__all__ = ['Endpoint', 'Header', 'Pet', 'TIMEOUT', 'WhistleAuthError',
           'WhistleClient', 'WhistleData', 'WhistleError']
